import axios from 'axios'
import get from 'lodash/get'
import sortBy from 'lodash/sortBy'
import uniqBy from 'lodash/uniqBy'
import qs from 'qs'

export const SERVER_TYPES = {
  gitlab: {
    getEntityType(category) {
      return category.entityType || 'issues'
    },
    getIssuesListUrl ({server, repository, entityType}) {
      return `${server.baseApiUrl}groups/${repository.groupId}/${entityType}`
    },
    getIssuesListParams ({server, repository, categoryFilters, additional, entityType}) {
      let serverTypeDefaults = {per_page: 30, state: 'opened', with_labels_details: true}
      let serverDefaults = get(server, 'params', {})
      let repositoryDefaults = get(repository, 'params', {})
      additional = additional || {}
      let concatenatedFilters = this.concatFilters([serverTypeDefaults,serverDefaults, repositoryDefaults, categoryFilters, additional])
      let combinedParams = {...serverTypeDefaults,...serverDefaults, ...repositoryDefaults, ...categoryFilters, ...additional, ...concatenatedFilters}
      if (entityType === 'merge_requests' && combinedParams.order_by === 'due_date') {
        combinedParams.order_by = 'created_at'
      }
      return combinedParams
    },
    concatFilters (filtersGroups) {
      let concatFields = ['labels', 'not[labels]']
      let values = {}
      concatFields.forEach(f => {
        filtersGroups.forEach(g => {
          if (g[f]) {
            values[f] = values[f] || []
            values[f].push(g[f])
          }
        })
        if (values[f]) {
          if (!f.startsWith('not[')) {
            values[f] = values[f].join(',')
          }
        }
      })
      return values


    },
    async getIssuesList ({server, repository, categoryFilters, params, category}) {
      let entityType = this.getEntityType(category)
      let query = this.getIssuesListParams({server, repository, categoryFilters, additional: params, entityType})
      let url = this.getIssuesListUrl({server, repository, category, entityType})
      let response = await axios.get(url, {
        params: query,
        paramsSerializer: function(params) {
          return qs.stringify(params, {arrayFormat: 'repeat'}
        )
     }})
      return this.parseIssuesListResponse({repository, response})
    },
    parseIssuesListResponse({repository, response}) {
      return {
        repository,
        objs: response.data,
        pagination: {
          current: response.headers['x-page'],
          next: response.headers['x-next-page'],
          prev: response.headers['x-prev-page'],
          totalPages: parseInt(response.headers['x-total-pages']),
          total: parseInt(response.headers['x-total'],)
        },
        status: response.status,
        response
      }
    },
    getSortValue (obj, sortKey) {
      if (sortKey === 'popularity') {
        return obj.upvotes
      }
      let value = obj[sortKey]
      if (sortKey === 'due_date') {
        value = value || ''
      }
      return value
    }
  }
}

export function getHandlers(s) {
  return SERVER_TYPES[s.type]
}
export async function mergeIssuesList(promises, options) {
  const data = {
    parsedResponses: [],
    objs: [],
    total: 0,
  }
  let sortedObjs = []
  for (var i = 0; i < promises.length; i++) {
    let promiseData = promises[i]
    let responseData = await promiseData.promise
    data.parsedResponses.push(responseData)
    data.total += responseData.pagination.total
    responseData.objs.forEach((o) => {
      o.repositoryObj = promiseData.repository
      o.categoryObj = promiseData.category
      o.uid = `${o.repositoryObj.baseApiUrl}/${o.iid}`
      let sortValue = getHandlers(o.repositoryObj.serverObj).getSortValue(o, options.orderBy)
      if (sortValue !== undefined) {
        sortedObjs.push({sortValue, obj: o})
      }
    })
  }
  sortedObjs = uniqBy(sortedObjs, 'obj.uid')
  // now, sorting everything, yay, fun
  sortedObjs = sortBy([...sortedObjs], ['sortValue'])
  if (options.sort === 'desc') {
    sortedObjs.reverse()
  }
  data.objs = sortedObjs.map((sortedObj) => {
    return sortedObj.obj
  })
  return data
}

export async function getIssues(config, params, taskSlugs) {
  let promises = []
  config.repositories.forEach((r) => {
    let categories = [...(r.categories || [r])]
    let matchingCategories = categories
    if (taskSlugs) {
      matchingCategories = matchingCategories.filter(r => {
        return taskSlugs.indexOf(r.taskSlug) > -1
      })
    }
    matchingCategories.forEach((c) => {
      let filters = c.filters || {}
      let promise = getHandlers(r.serverObj).getIssuesList({server: r.serverObj, repository: r, categoryFilters: filters, params, category: c})
      promises.push({repository: r, promise, category: c})
    })
  })
  let options = {
    sort:  params.sort || 'desc',
    orderBy:  params.order_by,
  }
  return await mergeIssuesList(promises, options)
}
